/*
 * MARTINS.js Free Edition
 * GPU-accelerated Augmented Reality for the web
 * Copyright (C) 2022  Alexandre Martins <alemartf(at)gmail.com>
 * https://github.com/alemart/martins-js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * view.ts
 * A viewer represents a virtual camera in 3D world space
 */

import Speedy from 'speedy-vision';
import { SpeedyMatrix } from 'speedy-vision/types/core/speedy-matrix';
import { CameraModel } from './camera-model';
import { Pose } from './pose';
import { ViewerPose } from './viewer-pose';
import { View, PerspectiveView } from './view';
import { StandardTransform } from './transform';
import { IllegalOperationError } from '../utils/errors';



/**
 * A viewer represents a virtual camera in 3D world space
 */
export class Viewer
{
    /** the pose of the viewer in 3D world space */
    private readonly _pose: ViewerPose;

    /** the views of this viewer (1 for monoscopic rendering; 2 for stereoscopic) */
    private readonly _views: View[];




    /**
     * Constructor
     * @param camera camera model
     */
    constructor(camera: CameraModel)
    {
        this._pose = new ViewerPose(camera);
        this._views = [ new PerspectiveView(camera) ];
    }

    /**
     * The pose of this viewer
     */
    get pose(): ViewerPose
    {
        return this._pose;
    }

    /**
     * The view of this viewer (only for monoscopic rendering)
     */
    get view(): View
    {
        /*
        if(this._views.length > 1)
            throw new IllegalOperationError('Use viewer.views for stereoscopic rendering');
        */

        return this._views[0];
    }

    /**
     * The views of this viewer
     */
    /*
    get views(): View[]
    {
        return this._views.concat([]);
    }
    */

    /**
     * Convert a pose from world space to viewer space
     * @param pose a pose in world space
     * @returns a pose in viewer space
     */
    convertToViewerSpace(pose: Pose): Pose
    {
        const modelMatrix = pose.transform.matrix;
        const viewMatrix = this._pose.viewMatrix;
        const modelViewMatrix = Speedy.Matrix(viewMatrix.times(modelMatrix));

        const transform = new StandardTransform(modelViewMatrix);
        return new Pose(transform);
    }
}